package com.css.simulation.resource.scheduler.scheduler;

import api.common.pojo.constants.DictConstants;
import api.common.pojo.dto.ProjectMessageDTO;
import api.common.util.CollectionUtil;
import api.common.util.JsonUtil;
import api.common.util.StringUtil;
import api.common.util.TimeUtil;
import com.css.simulation.resource.scheduler.consumer.ProjectConsumer;
import com.css.simulation.resource.scheduler.mapper.*;
import com.css.simulation.resource.scheduler.pojo.po.ClusterPO;
import com.css.simulation.resource.scheduler.pojo.po.ProjectPO;
import com.css.simulation.resource.scheduler.pojo.po.TaskPO;
import com.css.simulation.resource.scheduler.pojo.po.UserPO;
import com.css.simulation.resource.scheduler.pojo.to.KubernetesNodeTO;
import com.css.simulation.resource.scheduler.pojo.to.PrefixTO;
import com.css.simulation.resource.scheduler.service.TaskService;
import com.css.simulation.resource.scheduler.util.KubernetesUtil;
import com.css.simulation.resource.scheduler.util.ProjectUtil;
import io.kubernetes.client.openapi.ApiClient;
import lombok.SneakyThrows;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.data.redis.core.StringRedisTemplate;
import org.springframework.scheduling.annotation.Scheduled;
import org.springframework.stereotype.Component;

import javax.annotation.Resource;
import java.util.ArrayList;
import java.util.List;
import java.util.Set;

@Component
@Slf4j
public class ProjectScheduler {

    @Value("${scheduler.host.hostname}")
    String hostname;
    @Value("${scheduler.host.username}")
    String username;
    @Value("${scheduler.host.password}")
    String password;
    @Value("${scheduler.linux-path.job-yaml}")
    String jobYaml;
    // -------------------------------- Comment --------------------------------
    @Resource
    StringRedisTemplate stringRedisTemplate;
    @Resource
    TaskService taskService;
    @Resource
    TaskMapper taskMapper;
    @Resource
    ClusterMapper clusterMapper;
    @Resource
    ManualProjectMapper manualProjectMapper;
    @Resource
    AutoSubProjectMapper autoSubProjectMapper;
    @Resource
    ApiClient apiClient;
    @Resource
    ProjectConsumer projectConsumer;
    @Resource
    ProjectUtil projectUtil;
    @Resource
    UserMapper userMapper;


    /**
     * 调度项目启动
     */
    @Scheduled(fixedDelay = 60 * 1000)
    @SneakyThrows
    public void dispatchProject() {

        //1 查询已经在页面上点击运行的项目(后面会判断是排队中还是已经与运行了)
        List<ProjectPO> manualProjectList = manualProjectMapper.selectByNowRunState(DictConstants.PROJECT_RUNNING);
        List<ProjectPO> autoSubProjectList = autoSubProjectMapper.selectByNowRunState(DictConstants.PROJECT_RUNNING);
        List<ProjectPO> allProject = new ArrayList<>();
        allProject.addAll(manualProjectList);
        allProject.addAll(autoSubProjectList);


        for (ProjectPO project : allProject) {
            String projectId = project.getId();
            String projectType = project.getProjectType();
            long parallelism = Long.parseLong(project.getParallelism());
            String userId = project.getCreateUserId();
            UserPO userPO = userMapper.selectById(userId);
            String roleCode = userPO.getRoleCode();
            String useType = userPO.getUseType();
            ClusterPO clusterPO = null;
            String clusterId;
            boolean isSystem = false;
            if (DictConstants.ROLE_CODE_SYSADMIN.equals(roleCode) || DictConstants.ROLE_CODE_ADMIN.equals(roleCode)) {  //3-1 管理员账户和管理员子账户直接执行
                clusterId = DictConstants.SYSTEM_CLUSTER_ID;
                isSystem = true;
            } else if (DictConstants.ROLE_CODE_UESR.equals(roleCode)) { //3-2 普通账户，不管是独占还是共享，都在自己的集群里排队，根据自己的独占节点排队
                clusterPO = clusterMapper.selectByUserId(userId);
                if (clusterPO == null) {
                    log.error("ProjectScheduler--dispatchProject 项目 " + projectId + " 的创建用户 " + userId + " 没有分配集群！");
                    return;
                }
                clusterId = clusterPO.getId();
            } else if (DictConstants.ROLE_CODE_SUBUESR.equals(roleCode)) {
                if (DictConstants.USE_TYPE_EXCLUSIVE.equals(useType)) {   //3-3 普通子账户，根据自己的独占节点排队
                    clusterPO = clusterMapper.selectByUserId(userId);
                } else {    //3-4 共享子账户，根据父账户的共享节点排队
                    String parentUserId = userPO.getCreateUserId();
                    clusterPO = clusterMapper.selectByUserId(parentUserId);
                }
                if (clusterPO == null) {
                    log.error("ProjectScheduler--dispatchProject 项目 " + projectId + " 的创建用户 " + userId + " 没有分配集群！");
                    return;
                }
                clusterId = clusterPO.getId();
            } else {
                log.error("ProjectConsumer--dispatchProject 项目 " + projectId + " 的创建人 " + userId + " 为未知账户类型，不予执行！");
                return;
            }
            PrefixTO redisPrefix = projectUtil.getRedisPrefixByClusterIdAndProjectId(clusterId, projectId);
            // --------------------------------  判断项目是否已经在执行，如果执行则 continue --------------------------------
            if (StringUtil.isNotEmpty(stringRedisTemplate.opsForValue().get(redisPrefix.getProjectRunningKey()))) {
                continue;
            }
            // -------------------------------- 项目没有执行说明等待中 --------------------------------
            if (isSystem) { // 系统管理员直接执行
                run(clusterId, projectId, projectType, redisPrefix.getProjectWaitingKey(), redisPrefix.getProjectRunningKey(), parallelism);
                return;
            }
            int simulationLicenseNumber = clusterPO.getNumSimulationLicense();
            // 获取该用户正在运行的项目数量
            Set<String> clusterRunningKeySet = stringRedisTemplate.keys(redisPrefix.getClusterRunningPrefix() + "*");
            List<String> runningProjectSet = null;
            // cluster:${clusterId}:running:${projectId}
            if (CollectionUtil.isNotEmpty(clusterRunningKeySet)) {
                runningProjectSet = projectUtil.getRunningProjectList(clusterRunningKeySet);
                if (CollectionUtil.isNotEmpty(runningProjectSet)) {
                    log.info("ProjectScheduler--dispatchProject 运行中的项目的 key 有：" + runningProjectSet);
                    long parallelismSum = 0;
                    for (String runningProjectKey : runningProjectSet) {
                        parallelismSum += JsonUtil.jsonToBean(stringRedisTemplate.opsForValue().get(runningProjectKey), ProjectMessageDTO.class).getParallelism();
                    }
                    if (parallelismSum < simulationLicenseNumber) {
                        if (parallelismSum + parallelism < simulationLicenseNumber) {
                            run(clusterId, projectId, projectType, redisPrefix.getProjectWaitingKey(), redisPrefix.getProjectRunningKey(), parallelism);
                            return;
                        }
                    }
                }
            }
            if ((CollectionUtil.isEmpty(clusterRunningKeySet) || CollectionUtil.isEmpty(runningProjectSet)) && parallelism < simulationLicenseNumber) {
                run(clusterId, projectId, projectType, redisPrefix.getProjectWaitingKey(), redisPrefix.getProjectRunningKey(), parallelism);
            }
        }
    }


    public void run(String clusterId, String projectId, String projectType, String projectWaitingKey, String projectRunningKey, long parallelism) {
        String projectJson = stringRedisTemplate.opsForValue().get(projectWaitingKey);
        if (StringUtil.isEmpty(projectJson)) {
            log.error("ProjectScheduler--run 项目 " + projectId + " 的开始消息查询失败，key 为：" + projectWaitingKey);
            return;
        }
        //1 获取一个剩余可用并行度最大的节点
        KubernetesNodeTO maxParallelismPNodeTO = projectUtil.getMaxParallelismNode();
        String maxRestParallelismNode = maxParallelismPNodeTO.getName();
        Long maxRestParallelism = maxParallelismPNodeTO.getMaxParallelism();

        //2 判断剩余可用并行度是否大于项目并行度，否则加入扩充队列
        if (maxRestParallelism > parallelism) {
            log.info("ProjectConsumer--run 集群 " + clusterId + " 将项目 " + projectId + "在节点" + maxRestParallelismNode + " 执行！");
            projectConsumer.parseProject(projectId, projectType, projectJson, "cluster:" + clusterId, projectRunningKey, maxRestParallelismNode, parallelism);
        } else if (maxRestParallelism > 0) {
            log.info("ProjectConsumer--run 集群 " + clusterId + " 将项目 " + projectId + "在节点" + maxRestParallelismNode + " 执行！");
            projectConsumer.parseProject(projectId, projectType, projectJson, "cluster:" + clusterId, projectRunningKey, maxRestParallelismNode, maxRestParallelism);
        }
    }


    /**
     * 处理 pod 超时
     * 同时也可处理 pod 莫名关闭，因为关闭之后也会超时
     */
    @Scheduled(fixedDelay = 60 * 1000)
    public void taskTimeout() {
        long timeout = 2 * 60 * 1000L;
        List<TaskPO> executingTaskList = taskMapper.selectByRunState(DictConstants.TASK_RUNNING);
        if (executingTaskList != null && executingTaskList.size() > 0) {
            log.info("ProjectScheduler--taskTimeout 正在运行的任务有：" + executingTaskList);
            for (TaskPO task : executingTaskList) {
                String userId = task.getCreateUserId();
                String projectId = task.getPId();
                String taskId = task.getId();
                PrefixTO redisPrefix = projectUtil.getRedisPrefixByUserIdAndProjectIdAndTaksId(userId, projectId, taskId);
                // 获取心跳时间
                String tickTime = stringRedisTemplate.opsForValue().get(redisPrefix.getTaskTickKey());
                if (StringUtil.isEmpty(tickTime)) {
                    log.error(redisPrefix.getTaskTickKey() + "，该 key 的心跳时间为空！");
                    continue;
                }
                long lastTickTime = Long.parseLong(tickTime);
                // 如果心跳超时则更改任务状态为 Aborted
                if (TimeUtil.getNow() - lastTickTime > timeout) {
                    String podName = stringRedisTemplate.opsForValue().get(redisPrefix.getTaskPodKey());
                    taskService.taskState(taskId, DictConstants.TASK_ABORTED, podName);
                }
            }
        }
    }

    /**
     * 解决 pod 莫名全部关闭但是 job 还在的问题
     * 检查如果有 job 在运行但是 pod 全部关闭的情况，此时需要重启一下 job
     */
    @Scheduled(fixedDelay = 30 * 1000)
    @SneakyThrows
    public void projectCheck() {
        //1 查询出正在运行中的 project
        List<ProjectPO> projectIdList = manualProjectMapper.selectByNowRunState(DictConstants.PROJECT_RUNNING);
        //2 根据 projectId 获取 pod
        for (ProjectPO project : projectIdList) {
            String projectId = project.getId();
            String userId = project.getCreateUserId();
            PrefixTO redisPrefix = projectUtil.getRedisPrefixByUserIdAndProjectId(userId, projectId);
            String lastNowString = stringRedisTemplate.opsForValue().get(redisPrefix.getProjectCheckKey());
            // 获取该项目的 job 中正在运行的 pod 数量
            List<String> allJobList = KubernetesUtil.getJob(apiClient, "default");
            List<String> allPodList = KubernetesUtil.getPod(apiClient, "default");
            long jobNumber = allJobList.stream().filter(jobName -> jobName.contains(projectId)).count();
            long podNumber = allPodList.stream().filter(podName -> podName.contains(projectId)).count();
            // 如果没有检查过且 pod 列表为空，则正式开始检查，设置第一次检查时间
            if (StringUtil.isEmpty(lastNowString) && jobNumber == 1L && podNumber == 0L) {
                log.info("ProjectScheduler--projectCheck 开始检查项目 " + projectId);
                stringRedisTemplate.opsForValue().set(redisPrefix.getProjectCheckKey(), TimeUtil.getNowString());
                return;
            }
            log.info("ProjectScheduler--projectCheck kubernetes 的命名空间 default 中正在运行的 pod 有：" + allPodList + "，其中项目 " + projectId + " 的任务个数为：" + podNumber);
            //  如果两次检查时间超过了 2 分钟，且仍然没有 pod 执行，则准备重启
            if (StringUtil.isNotEmpty(lastNowString) && jobNumber == 1L && podNumber == 0L && Long.parseLong(TimeUtil.getNowString()) - Long.parseLong(lastNowString) > (long) 120 * 1000) {
                // 删除检查
                stringRedisTemplate.delete(redisPrefix.getProjectCheckKey());
                try {
                    // 删除 job
                    KubernetesUtil.deleteJob(apiClient, "default", "project-" + projectId);
                } catch (Exception e) {
                    log.info("ProjectScheduler--projectCheck 删除项目 " + projectId + " 的 job 失败，可能是已经删除！");
                }
                log.info("ProjectScheduler--projectCheck 重新执行项目" + projectId);
                KubernetesUtil.applyYaml(hostname, username, password, jobYaml + "project-" + projectId + ".yaml");
            }
        }
    }
}
